#!/usr/bin/perl

# Author : Damien KROTKINE (damien@tuxfamily.org)
#
# Contributors : 
# Charles LONGEAU (chl@tuxfamily.org)
#
# Copyright (C) 2002 Damien KROTKINE (damien@tuxfamily.org)
# Copyright (C) 2002 Charles LONGEAU (chl@tuxfamily.org)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

package Libconf;
use strict;
use vars qw(@ISA %EXPORT_TAGS @EXPORT_OK $VERSION $INDENT_SPACES $TEMPLATES_DIRECTORY %templates );

@ISA = qw(Exporter);
%EXPORT_TAGS = (
    functions => [ qw() ],
    settings => [ qw() ],
);
$EXPORT_TAGS{all} = [ map { @$_ } values %EXPORT_TAGS ];
@EXPORT_OK = map { @$_ } values %EXPORT_TAGS;

$VERSION='0.22';

$INDENT_SPACES = '    ';

sub new {
    my ($pkg, $filename, $template_name, $validation_name, $template_filename, $safe_write) = @_;
    -e $filename or die ("$filename not found");
    $template_filename ?
      require $template_filename : eval "require Libconf::templates::$template_name" . '_template';
    my %out;
    $out{filename} = $filename;
    $out{template_name} = $template_name;
    $validation_name and $out{validation_name} = $validation_name;
    $out{atoms} = [];
    $safe_write and $out{timestamp} = getTimestamp($filename);
    readConf(\%out);
    bless \%out, $pkg;
    \%out;
}

sub getTimestamp {
    my ($filename) = @_;
    my @stat;
    @stat = stat($filename);
    return($stat[9]);
}

sub readConf {
    my ($out) = @_;
    my $filename = $out->{filename};
    -e $filename or die ("$filename not found");
    my @file = cat_($filename);
    parse($out, \@file);
}

sub clear {
    my ($out) = @_;
    $out->{atoms} = [];
}

sub size {
    my ($out) = @_;
    scalar (@{$out->{atoms}});
}

# COMMENTS TO BE REVIEWED
# $out : the structure
# $index : the atom index. If -1, then the template will try to find the matching atom
# $ref_args : a reference on a hash containing the datas
# returns : 
#   0 if ok,
#   -1 if $index bigger than @atoms,
#   -2 if the template didn't succeed finding the mathing atom ($index was -1),
#   -3 if there is an error setting the values in the atom
sub editAtom {
    my ($out, $index, $ref_args) = @_;
    my %args=%$ref_args;
    eval($templates{$out->{template_name}}{edit_atom});
    $@ and die $@;
#    validate($out, @{$out->{atoms}}[$index]);
}

sub getAtom {
    my ($out, $index) = @_;
    $out->{atoms}[$index];
}

sub findAtomPos {
    my ($out, $ref_args, $first_atom, $last_atom) = @_;
    my %args=%$ref_args;
    eval($templates{$out->{template_name}}{find_atom_pos});
}

sub appendAtom {
    my ($out, $ref_args) = @_;
    my %args = %$ref_args;
    my $index = @{$out->{atoms}};
    push(@{$out->{atoms}}, {});
    eval($templates{$out->{template_name}}{edit_atom});
#    validate($out, @{$out->{atoms}}[$index]);
}

sub insertAtom {
    my ($out, $index, $ref_args) = @_;
    $index >= @{$out->{atoms}} and return appendAtom($out, $ref_args);
    my %args = %$ref_args;
    splice(@{$out->{atoms}}, $index, 0, {});
    $args{sections} ||= @{$out->{atoms}}[$index-1]->{sections};
    $args{sections} ||= [];
    eval($templates{$out->{template_name}}{edit_atom});
#    validate($out, @{$out->{atoms}}[$index]);
}

sub deleteAtom {
    my ($out, $index) = @_;  
    splice(@{$out->{atoms}}, $index, 1);
#    validate($out, @{$out->{atoms}}[$index]);
}

sub compare_depth {
    my ($ref1, $ref2) = @_;
    my @sections1 = @$ref1;
    my @sections2 = @$ref2;
    @sections1 == @sections2 or return 0;
    foreach my $i (0..@sections1-1) {
        @sections2[$i]->{name} eq @sections1[$i]->{name} or return 0;
    }
    1;
}

sub deleteSection {
    my ($out, $index, $no_end_section) = @_;
    my $atom = $out->getAtom($index);
    #$atom->sections;
    my @end_positions = $no_end_section ?
      $out->findAtomPos( { type => 'SECTION' }) :
        $out->findAtomPos( { type => 'ENDSECTION', sections => $atom->{sections} });
    my $end;
    foreach (@end_positions) {
        $_ > $index and $end = $_, last;
    }
    !defined $end && $no_end_section and $end = @{$out->{atoms}}-1;
    defined $end or die "no ENDSECTION for SECTION n $index";
    splice(@{$out->{atoms}}, $index, $end-$index+1);
}

sub clearSection {
    my ($out, $index, $no_end_section) = @_;
    my $atom = $out->getAtom($index);
    #$atom->sections;
    my @end_positions = $no_end_section ?
      $out->findAtomPos( { type => 'SECTION' }) :
        $out->findAtomPos( { type => 'ENDSECTION', sections => $atom->{sections} });
    my $end;
    foreach (@end_positions) {
        $_ > $index and $end = $_, last;
    }
    !defined $end && $no_end_section and $end = @{$out->{atoms}}-1;
    print "plop\n";
    defined $end or die "no ENDSECTION for SECTION n $index";
    splice(@{$out->{atoms}}, $index+1, $end-$index);
}

sub writeConf {
    my ($out, $filename) = @_;
    $filename ||= $out->{filename};
    my $template_name = $out->{template_name};
    my @out_atoms = @{$out->{atoms}};
    if (defined $out->{timestamp} && (getTimestamp($filename) != $out->{timestamp})) {
        print "Warning $filename has changed !\n";
    }
    local *F;
    open F, "> $filename" or die "cannot create config file $filename (do you have proper permissions?)";
    foreach my $atom (@out_atoms) {
#	print F "\n" x $$atom{line_feed};
#	$$atom{comments} and print F eval($templates{$template_name}{comment_output}) foreach split("\n", $$atom{comments});
        my $output_text = '';
        my $output_indentation = $out->{current_indentation};
        if (exists $templates{$template_name}->{output}{$atom->{type}}) {
            eval($templates{$template_name}->{output}{$atom->{type}});
        } else {
            if (defined($atom->{type})) {
                print F '## WARNING - No valid $template->{output}{' . $atom->{type} . "}\n";
                die '## WARNING - No valid $template->{output}{' . $atom->{type} . "}\n";
            }
        }
        print F handle_output($output_text, $output_indentation, $atom->{comments});
    }
}

sub parse {
    my ($out, $file) = @_;
    my @file = @$file;
    my $item = 0;
#    my $next_item = 0;
    my $template_name = $out->{template_name};
    my @out_atoms;
    my $matched;
    my $keep_atom;
    my $switch_to_next_atom = 0;
    my $switch_to_prev_atom = 0;

    $out->{current_sections} ||= [];
    my $line_idx = 1;
    foreach my $in (@file) {
	$out_atoms[$item] ||= {};
        $out_atoms[$item]->{comments} ||= [];
	$in =~ /^\s*\n$/ and push(@{$out_atoms[$item]->{comments}}, { comment => '', inline => 0 }), next;
#	$in eq "\n" and $out_atoms[$item]->{line_feed}++, next;
	chomp $in;
	#see if we are in a line previously commented
	if ($out_atoms[$item]->{current_comment}) {
	    my $c = $out_atoms[$item]->{current_comment};
	    if ($in =~ s/^(.*\Q$c\E)//) {
                push(@{$out_atoms[$item]->{comments}}, { comment => $1, inline => $in =~ /^\s*$/ ? 0 : -1 });
		undef $out_atoms[$item]->{current_comment};
	    } else {
                push(@{$out_atoms[$item]->{comments}}, { comment => $in, inline => 0 });
		$in = '';
	    }
	} else {
	    # we remove comments
	    # the comment structure : [ open_char, close_char, nested ]
	    my ($open, $close, $nested);
	    foreach (@{$templates{$template_name}{comments}}) {
		($open, $close, $nested) = @$_;
		if ($nested) {
		    my ($level, $start) = (0, 0);  #we don't use $out_atoms[$item]->{current_level} directly to avoid overriding it.
		    my $end = 0; #length($in);
		    for (my $i=0; $i<length($in); $i++) {
			substr($in, $i, length($open)) eq $open and $level++, $start ||= $i;
			if ($level && substr($in, $i, length($close)) eq $close) {
			    $level--;
			    $level or $end = $i-$start;
			}
		    }
		    $end and push(@{$out_atoms[$item]->{comments}}, { comment => substr($in, $start, $end + length($close)),
                                                                    inline => $start > 0 ? 1 : 0 });
		    substr($in, $start, $end + length($close)) = '';
		    $level and $out_atoms[$item]->{current_comment} = $close, $out_atoms[$item]->{current_level} = $level;
		} elsif (defined $open && defined $close) {
		    $in =~ s/(\Q$open\E.*\Q$close\E)/push(@{$out_atoms[$item]->{comments}}, { comment => $1, inline => length($') ? -1 : (length($`) ? 1 : 0) }); undef/eg;
		    #$in =~ s/(\Q$open\E.*\Q$close\E)/$out_atoms[$item]->{comments} .= "$1\n"; undef/eg;
		    $in =~ s/(\Q$open\E.*)$// and push(@{$out_atoms[$item]->{comments}}, { comment => $1,
                                                                                        inline => $in =~ /^\s*$/ ? 0 : 1 }
                                                      ), $out_atoms[$item]->{current_comment} = $close, last;
		} elsif (defined $open) {
                    if ($in =~ s/(\Q$open\E.*)$//) {
                        my $tmp = $1;
                        push(@{$out_atoms[$item]->{comments}}, { comment => $tmp,
                                                                 inline => $in =~ /^\s*$/ ? 0 : 1 });
                    }
		}
	    }
        }
	### End Comment handling ###
        $matched = 0;
        $keep_atom = 0;
        $switch_to_next_atom = 0;
        $switch_to_prev_atom = 0;
        my $atom = $out_atoms[$item];
        foreach (@{$templates{$template_name}{rules}}) {
	    eval ( $_ );
	    $@ and die qq(error while applying the rule '$_' to '$in' line n $line_idx : $@);
            # the last rule asked to go to next atom
            if ($switch_to_next_atom) {
                $item++;
                $out_atoms[$item] ||= {};
                $out_atoms[$item]->{comments} ||= [];
                $atom = $out_atoms[$item];
                $switch_to_next_atom;
            }
            if ($switch_to_prev_atom) {
                $item--;
                $atom = $out_atoms[$item];
                $switch_to_prev_atom = 0;
            }
#            $next_item = $matched && !$keep_atom;
	    $matched and last;
	}
#	validate($out, $out_atoms[$item]);
        if ($in !~ /^\s*$/){
            push @{$out_atoms[$item]->{invalid}}, $in;
            print "\nLibconf.pm warning: can't parse line n" . $line_idx . " :\n$in\n";
        }
#       $keep_atom or $item++;
        $matched && !$keep_atom and $item++;
#	if (!$keep_atom) {
#	    $item++;
#	    $next_item = 0;
#	}
    } continue {
        $line_idx++;
    }
    $out->{atoms} = [@out_atoms];
}

sub handle_output {
    my ($text, $indentation, $comments) = @_;
    my $ret = '';
    my $flag = 1;
    $indentation = $INDENT_SPACES x $indentation;
    foreach my $comment (@$comments) {
        $comment->{inline} == -1 and $ret .= $indentation . $comment->{comment} . " $text\n", $flag = 0;
        $comment->{inline} == 0 and $ret .= $indentation . $comment->{comment} . "\n";
        $comment->{inline} == 1 and $ret .= $indentation . "$text " . $comment->{comment} . "\n", $flag = 0;
    }
    $flag && length($text) and $ret .= $indentation . $text . "\n";
    $ret;
}

#validation stuff

sub setUniq {
    my ($obj) = @_;
    my @to_delete;
    my $i = 0;
    foreach my $atom (@{$obj->{atoms}}) {
        my %atom2 = %$atom;
        delete @atom2{qw(comments value list)};
        my @pos = findAtomPos($obj, \%atom2);
        if (@pos > 1 && $pos[-1] != $i) {
            push @to_delete, $i;
        }
        $i++;
    }
    print "@to_delete\n";
    deleteAtom ($obj, $_) foreach (reverse @to_delete);
}

sub debug {
    $ENV{DEBUG} or return;
    my $i = 0;
    while (caller($i)) { $i++ }
    my ($package, $filename, $line, $subroutine, $hasargs, $wantarray, $evaltext, $is_require, $hints, $bitmask) = caller(2);
    print '   ' x ($i-3);
    print "\033[35m";
    print $subroutine . "\t\033[36m", @_;
    print "\033[0m\n";
}

# sub validate {
#     my ($out, $atom) = @_;
#     $out->{validation_name} and eval($validation{$out->{validation_name}}->{rules});
# }

sub cat_ { local *F; open F, $_[0] or return; my @l = <F>; wantarray ? @l : join '', @l }
sub output { my $f = shift; local *F; open F, ">$f" or die "output in file $f failed: $!\n"; print F foreach @_; 1 }
sub member { my $e = shift; foreach (@_) { $e eq $_ and return 1 } 0 }
sub map_index(&@) {
    my $f = shift;
    my @v; local $::i = 0;
    map { @v = &$f($::i); $::i++; @v } @_;
}
sub if_($@) {
    my $b = shift;
    $b or return ();
    wantarray() || @_ <= 1 or die("if_ called in scalar context with more than one argument " . join(":", caller()));
    wantarray() ? @_ : $_[0];
}
sub fold_left(&@) {
    my ($f, $initial, @l) = @_;
    local ($::a, $::b);
    $::a = $initial;
    foreach (@l) { $::b = $_; $::a = &$f() }
    $::a
}
sub map_index(&@) {
    my $f = shift;
    my @v; local $::i = 0;
    map { @v = &$f($::i); $::i++; @v } @_;
}

1;
